
//		Say.cpp - Simple SAPI plugin for Take Command


#define UNICODE						1
#define _UNICODE					1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>
#include <sapi.h>

#pragma warning(push)
#pragma warning(disable: 4995)
#include <sphelper.h>
#pragma warning(pop)

#include "PlugIn.h"
#include "TakeCmd.h"


#define VER_MAJOR					1
#define VER_MINOR					2
#define VER_BUILD					1
#define VER_PATCH					0

#define PLUGIN_NAME					L"Say"

#ifdef _WIN64
#define PLATFORM					L"x64"
#else
#define PLATFORM					L"x86"
#endif




#define MAX_ARG_LEN					8192
#define MAX_ARGS					1024
#define SHORT_BUF_LEN				1024
#define BUFFER_SIZE					8192
#define MAX_FILENAME_LEN			32768

#define ERRHANDLE					GetStdHandle( STD_ERROR_HANDLE )

#define FILE_OEM                    0
#define FILE_UTF16                  1
#define FILE_UTF8                   2

#define PILCROW						0xb6
#define SECTION_MARK				0xa7
#define INVALID_CHAR				'?'
#define TAB							0x09
#define SPACE						0x20
#define CR							0x0d
#define LF							0x0a


LPTSTR arg[MAX_ARGS];
int arglen[MAX_ARGS];
int numargs = 0;
wchar_t allargs[MAX_ARG_LEN]		= L"";



const wchar_t SayHelpText[] = L"Speak the specified text.\n\
\n\
SAY /A /D /H /K /L /N /X text @filename  \r\n\
\n\
\t/A  speak command-line text Asynchronously\n\
\t/D  Disable SAPI XML tags (default)\n\
\t/H  speak Help text\n\
\t/K  restart voice; cancel any asynchronous speech  \n\
\t/L  Line breaks in text files are significant\n\
\t/N  speak line Numbers in text files\n\
\t/X  enable SAPI XML tags\n\
\n\
Put options before any text or filenames.  Text and files  \n\
will be read in the order they are listed on the command\n\
line.  Text can also be piped or redirected into SAY;\n\
redirected input will be read after any text or files\n\
specified on the command line.\n";


const wchar_t SaySpokenHelp[] = L"The SAY command speaks the specified text. \
Valid arguments include: options, text to be spoken, and file names. Precede \
any option with a forward slash. Precede any filename with a commercial at sign. \
Slash <pron sym=\"ey\"/> speaks command-line text asynchronously. \
Slash D disables XML tags. \
Slash H speaks this help text. \
Slash K restarts the voice and cancels any asynchronous speech. \
Slash L means that line breaks in text files are significant. \
Slash N speaks line numbers in text files. \
Slash X enables XML tags. \
Text and files will be read in the order they are listed on the command line. \
Text can also be piped or redirected into SAY. Redirected input will be read \
after any text or files from the command line.";


const wchar_t VoiceHelpText[] = L"List available voices, or select the voice for SAY to use.\n\
\n\
VOICE /A /D /H name  \r\n\
\n\
\t/A  read the list Aloud\n\
\t/D  select the system Default voice\n\
\t/H  speak Help text\n\
\n\
Put any options before the voice name.  If no argument is given,  \n\
VOICE lists all available voices.  \n";


const wchar_t VoiceSpokenHelp[] = L"The VOICE command lists available voices, or selects the voice to use. \
Valid arguments include options, or the name of the voice to select. Precede any option with a forward slash. \
Slash <pron sym=\"ey\"/> reads the list aloud. \
Slash D selects the system default voice. \
Slash H speaks this help text. \
Any options must come before the voice name. You may use wild cards in the voice name. ";


const wchar_t _NumVoicesHelpText[] = L"_NUMVOICES  \n\
\n\
Returns the number of voices available, or -1  \n\
on any error.\n";


const wchar_t _VoiceHelpText[] = L"_VOICE  \n\
\n\
Returns the name of the current voice, or (unknown)  \n\
on any error.\n";


const wchar_t _VoiceNumHelpText[] = L"_VOICENUM  \n\
\n\
Returns the index of the current voice, 0 to _NUMVOICES - 1;  \n\
or -1 on any error.\n";


const wchar_t F_VoiceHelpText[] = L"@VOICE[n]  \n\
\n\
Returns the name of a voice.  n is an integer from 0 to  \n\
%_NUMVOICES - 1.  Returns -1 if n is too high.\n";



ISpVoice *MyVoice						= NULL;
ISpObjectToken *UserVoice				= NULL;


wchar_t FilenameSave[MAX_FILENAME_LEN]	= L"";
HANDLE FileHandle						= NULL;
BOOL FileEOF							= FALSE;
int FileEncoding						= FILE_OEM;
BYTE *FileBuffer						= NULL;
DWORD BytesRead							= 0;
int BufferIndex							= 0;

BYTE *ClipboardData						= NULL;
SIZE_T ClipboardSize					= 0;
size_t ClipPtr							= 0;

int CmdUtterance						= 0;


OSVERSIONINFOEXW OSVerInfo;



const wchar_t PluginFeatures[]	= L"say,voice,_numvoices,_voice,_voicenum,@voice";

const wchar_t HelpWindowTitle[]	= L"Say Plugin Help";



BOOL PipeShell()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"%[_transient]%[_pipe]";


	ExpandVariables( temp, 1 );

	if ( wcscmp( temp, L"00" ) )
		return TRUE;
	else
		return FALSE;
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/say.html";
	piInfo.pszDescription	= (LPTSTR) L"Simple SAPI speech support";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



int CreateVoice()
{
	HRESULT rv = S_OK;

	if ( MyVoice != NULL )
		return 0;
	else {
		rv = CoCreateInstance( CLSID_SpVoice, NULL, CLSCTX_ALL, IID_ISpVoice, (void **) &MyVoice );

		if ( MyVoice && UserVoice )
			MyVoice->SetVoice( UserVoice );

		return rv;
	}
}



int ReleaseVoice()
{
	if ( MyVoice != NULL ) {
		MyVoice->Release();
		MyVoice = NULL;
	}

	return 0;
}



int RestartVoice()
{
	ReleaseVoice();
	Sleep( 100 );
	return CreateVoice();
}



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( GetStdHandle(STD_ERROR_HANDLE), shortbuf );

	return 0;
}



BOOL GetVersion2( OSVERSIONINFOEXW *VerInfo )
{
	if ( !VerInfo )
		return -1;


	typedef NTSTATUS ( WINAPI _RtlGetVersionProc ) ( OSVERSIONINFOEXW *VerInfo );
	HMODULE NtDllHandle						= GetModuleHandle( L"NtDll.dll" );
	_RtlGetVersionProc *RtlGetVersionProc	= NULL;


	memset( VerInfo, 0x00, sizeof( OSVERSIONINFOEXW ) );
	VerInfo->dwOSVersionInfoSize = sizeof( OSVERSIONINFOEXW );

	if ( NtDllHandle )
		RtlGetVersionProc = (_RtlGetVersionProc*) GetProcAddress( NtDllHandle, "RtlGetVersion" );

	if ( RtlGetVersionProc ) {
		RtlGetVersionProc( VerInfo );
		if ( VerInfo->dwMajorVersion >= 6 )
			return TRUE;
	}

	memset( VerInfo, 0x00, sizeof( OSVERSIONINFOEXW ) );
	VerInfo->dwOSVersionInfoSize = sizeof( OSVERSIONINFOEXW );

	return ( GetVersionEx( (OSVERSIONINFOW*) VerInfo ) != 0 );
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	GetVersion2( &OSVerInfo );

	if ( ( OSVerInfo.dwMajorVersion < 5 ) || ( ( OSVerInfo.dwMajorVersion == 5 ) && ( OSVerInfo.dwMinorVersion == 0 ) ) ) {
		if ( !PipeShell() ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Windows XP or later required! \r\n" );
			Qprintf( ERRHANDLE, L"Plugin not loaded. \r\n" );
		}

		return -1;
	}

	if ( CreateVoice() ) {
		if ( !PipeShell() ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error initializing voice! \r\n" );
			Qprintf( ERRHANDLE, L"Plugin not loaded. \r\n" );
		}

		return -1;
	}

	wchar_t shortbuf[MAX_ARG_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s plugin v%d.%d.%d loaded. \r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	ReleaseVoice();

	if ( UserVoice )
		UserVoice->Release();

	if ( FileBuffer )
		FreeMem( FileBuffer );

	if ( ClipboardData )
		free( ClipboardData );

	return 0;
}



#define PARSE_BREAK_SPACES		1
#define PARSE_BREAK_COMMAS		2
#define PARSE_SLASHES_KLUDGE	4
#define	PARSE_QUOTES_KLUDGE		8
#define PARSE_EQUALS_KLUDGE		16
#define PARSE_ONE_ARG_KLUDGE	32
#define PARSE_RETAIN_QUOTES		64
#define PARSE_FORCE_UPPERCASE	128
#define PARSE_NO_TRIM			256
#define PARSE_IGNORE_QUOTES		512
#define PARSE_EQ_QT_KLUDGE		1024


int ParseArgs( LPTSTR inString, unsigned int flags )
{
	/*
		flags:  bit 0 - break args at spaces
				bit 1 - break args at commas (either bit 0 or bit 1 must be set)
				bit 2 - slashes kludge:   treat /a/b like /a /b
				bit 3 - quotes kludge:    treat /a"foo" like /a:"foo"
				bit 4 - equals kludge:    break at first unquoted equals sign
				bit 5 - one-arg kludge:   if arg does not begin with /, allow unquoted spaces
				bit 6 - don't swallow double quotes
				bit 7 - force uppercase
				bit 8 - don't trim spaces from the right
				bit 9 - disable all special handling of double quotes
	*/

	if ( !inString )
		return -1;

	numargs = 0;

	int argnum = 0;
	int i = 0;
	int a = 0;
	int b = 0;
	wchar_t ch = '\0';
	BOOL quote  = FALSE;
	BOOL adv    = FALSE;
	BOOL done   = FALSE;
	int eqkl    = 0;

	if ( ( flags & ( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_EQUALS_KLUDGE ) ) == 0 )
		flags |= PARSE_BREAK_SPACES;

	for ( i = 0; i < MAX_ARGS; i++ ) {
		arg[i] = NULL;
		arglen[i] = 0;
	}

	for ( i = 0; !done; i++ ) {
		ch = inString[i];
		adv = FALSE;

		if ( ( ch == '"' ) && ( ( flags & PARSE_IGNORE_QUOTES ) == 0 ) ) {
			quote = !quote;

			if ( quote && ( a == 2 ) && ( arg[argnum][0] == '/' ) && ( iswalpha( arg[argnum][1] ) ) && ( flags & PARSE_QUOTES_KLUDGE ) ) {
				if ( flags & PARSE_RETAIN_QUOTES ) {
					if ( b < ( MAX_ARG_LEN - 1 ) )
						allargs[b++] = ':';
				}
				else
					ch = ':';
			}
			else if ( ( flags & PARSE_RETAIN_QUOTES ) == 0 )
				continue;
		}

		if ( iswspace( ch ) && ( a == 0 ) && !quote )
			continue;

		if ( ( a == 0 ) && ( ch != '/' ) && ( ( flags & ( PARSE_BREAK_SPACES | PARSE_ONE_ARG_KLUDGE ) ) == ( PARSE_BREAK_SPACES | PARSE_ONE_ARG_KLUDGE ) ) )
			flags &= ~PARSE_BREAK_SPACES;

		if ( iswspace( ch ) && !quote && ( a == 1 ) && ( argnum > 0 ) && ( flags & PARSE_EQUALS_KLUDGE ) && ( arg[argnum][0] == '=' ) )
			continue;

		if ( iswspace( ch ) && !quote && ( flags & PARSE_BREAK_SPACES ) )
			adv = TRUE;

		if ( ( ch == ',' ) && !quote && ( flags & PARSE_BREAK_COMMAS ) )
			adv = TRUE;

		if ( ( ch == '/' ) && !quote && ( a > 1 ) && ( arg[argnum][0] == '/' ) && ( flags & PARSE_SLASHES_KLUDGE ) ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
		}

		if ( ( ch == '=' ) && !quote && ( eqkl == 0 ) && ( a > 0 ) && ( arg[argnum][0] != '/' ) && ( flags & PARSE_EQUALS_KLUDGE ) ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
			eqkl = argnum;
			flags &= ~( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS );
			//- Printf( L"   * equals kludge invoked; eqkl = %d\n", eqkl );

			if ( flags & PARSE_EQ_QT_KLUDGE )
				flags |= PARSE_IGNORE_QUOTES;
		}
		else if ( ( ch == '=' ) && !quote && ( flags & PARSE_EQUALS_KLUDGE ) ) {
			eqkl = -1;
			flags &= ~( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS );

			if ( flags & PARSE_EQ_QT_KLUDGE )
				flags |= PARSE_IGNORE_QUOTES;
		}

		if ( ch == '\0' ) {
			if ( a )
				argnum++;

			done = TRUE;
			continue;
		}

		if ( adv ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
			quote = FALSE;
			continue;
		}

		if ( b >= ( MAX_ARG_LEN - 1 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Buffer overflow: b = %d, MAX_ARG_LEN = %d \r\n", b, MAX_ARG_LEN );
			return -1;
		}

		if ( argnum < MAX_ARGS ) {
			if ( a == 0 )
				arg[argnum] = allargs + b;

			if ( flags & PARSE_FORCE_UPPERCASE )
				ch = towupper( ch );

			a++;
			arglen[argnum]  = a;
			allargs[b++] = ch;
			allargs[b]   = '\0';
		}

	}

	if ( ( flags & PARSE_NO_TRIM ) == 0 ) {
		for ( i = 0; i < argnum; i++ ) {
			while ( ( arglen[i] > 0 ) && iswspace( arg[i][arglen[i] - 1] ) ) {
				arglen[i]--;
				arg[i][arglen[i]] = '\0';
			}
		}
	}

	numargs = argnum;
	return argnum;
}



BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	unsigned int value = 0;
	wchar_t *sep1 = NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( !EndChar && ( *sep1 != '\0' ) )
			return -1;

		*outValue = value;

		if ( EndChar != NULL )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( !EndChar && ( *sep1 != '\0' ) )
			return -1;

		*outValue = value;

		if ( EndChar != NULL )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



int IsVowel( wchar_t ch )
{
	wchar_t upper = towupper(ch);

	if ( ( upper == 'A' ) || ( upper == 'E' ) || ( upper == 'I' ) || ( upper == 'O' ) || ( upper == 'U' ) || ( upper == 'E' ) || ( upper == 'Y' ) )
		return 1;
	else
		return 0;
}



int BreakCheck( )
{
	try {
		QueryKeyWaiting();
	}
	catch ( ... ) {
		return -3;
	}

	return 0;
}



int GetErrorText( int err, LPTSTR outString )
{
	if ( !outString )
		return -1;

	if ( err == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error" );
	else if ( FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, err, 0, outString, SHORT_BUF_LEN, NULL ) == 0 )
		wsprintf( outString, L"Error %d", err );

	size_t l = wcslen( outString );

	if ( l > 0 )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



int GetByte( )
{
	if ( !FileHandle )
		return -1;

	wchar_t ErrorText[SHORT_BUF_LEN] = L"";
	int err = 0;

	if ( BufferIndex >= (int) BytesRead ) {

		if ( FileEOF ) {
			BufferIndex = 0;
			return -1;
		}
		else if ( ReadFile( FileHandle, FileBuffer, BUFFER_SIZE, &BytesRead, NULL ) == 0 ) {
			err = GetLastError();

			if ( err != ERROR_BROKEN_PIPE ) {
				GetErrorText( err, ErrorText );
				DisplayErrorHeader();
				Qprintf( GetStdHandle( STD_ERROR_HANDLE), L"Error reading \"%s\" (2): %s\r\n", FilenameSave, ErrorText );
			}

			BytesRead = BufferIndex = 0;
			FileEOF = TRUE;
			return -1;
		}
		else if ( BytesRead == 0 ) {
			BufferIndex = 0;
			FileEOF = TRUE;
			return -1;
		}

		BufferIndex = 0;
	}

	return FileBuffer[BufferIndex++];
}



int GetChar( )
{
	int b1 = 0, b2 = 0, b3 = 0;
	wchar_t ch = '\0';


	if ( ClipboardData ) {
		if ( ClipPtr >= ClipboardSize ) {
			FileEOF = TRUE;
			return 0;
		}

		b1 = ClipboardData[ClipPtr] | ( ClipboardData[ClipPtr + 1] << 8 );
		ClipPtr += 2;

		if ( b1 == '\0' )
			FileEOF = TRUE;

		return b1;
	}



	b1 = GetByte();
	if ( b1 < 0 )
		return b1;

	if ( FileEncoding == FILE_UTF16 ) {
		b2 = GetByte();
		if ( b2 < 0 )
			return b2;

		ch = b1 | ( b2 << 8 );
	}
	else if ( FileEncoding == FILE_UTF8 ) {

		if ( b1 < 0x80 )
			ch = b1;
		else if ( b1 < 0xc0 )
			ch = INVALID_CHAR;
		else if ( b1 < 0xe0 ) {
			b2 = GetByte();
			if ( b2 < 0 )
				return b2;

			if ( ( b2 & 0xc0 ) != 0x80 )
				ch = INVALID_CHAR;
			else
				ch = ( ( b1 & 0x1f ) << 6 ) | ( b2 & 0x3f );
		}
		else if ( b1 < 0xf0 ) {
			b2 = GetByte();
			if ( b2 < 0 )
				return b2;

			b3 = GetByte();
			if ( b3 < 0 )
				return b3;

			if ( ( ( b2 & 0xc0 ) != 0x80 ) || ( ( b3 & 0xc0 ) != 0x80 ) )
				ch = INVALID_CHAR;
			else
				ch = ( ( b1 & 0x0f ) << 12 ) | ( ( b2 & 0x3f ) << 6 ) | ( b3 & 0x3f );
		}
		else
			ch = INVALID_CHAR;

	}
	else
		ch = b1;

	return ch;
}



int SpeakText( LPTSTR Filename, BOOL Verbose, BOOL SayLineNumbers, BOOL EnableXML, BOOL Async, BOOL BreakOnEOL, LPTSTR LiteralText )
{

	wchar_t OutBuffer[MAX_ARG_LEN] = L"";
	wchar_t ErrorText[SHORT_BUF_LEN] = L"";
	wchar_t LineNumber[SHORT_BUF_LEN] = L"";
	int i = 0, j = 0, utterance = 0;
	int ch = '\0';
	wchar_t lastch = '\0';
	int crcount = 0, lfcount = 0, dotscount = 0, vowelscount = 0;
	BOOL skipme = FALSE;
	BOOL CloseMe = FALSE;

	DWORD SpeakFlags = EnableXML ? SPF_IS_XML : SPF_IS_NOT_XML;

	if ( Async && ( LiteralText != NULL ) )
		SpeakFlags |= SPF_ASYNC;

	FileEOF = FALSE;
	FileEncoding = FILE_OEM;
	BufferIndex = 0;
	BytesRead = 0;

	HANDLE ClipHandle		= NULL;
	SIZE_T ClipSize			= 0;
	BYTE *ClipTmp			= NULL;



	if ( ClipboardData ) {
		free( ClipboardData );
		ClipboardData	= NULL;
		ClipboardSize	= 0;
		ClipPtr			= 0;
	}


	if ( !LiteralText ) {

		if ( !FileBuffer ) {
			FileBuffer = (BYTE*) AllocMem( BUFFER_SIZE );
			if ( !FileBuffer ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Error allocating file buffer\r\n" );
				return -1;
			}
		}

		if ( !wcscmp( Filename, L"-" ) ) {
			FileHandle = GetStdHandle( STD_INPUT_HANDLE );
			CloseMe = FALSE;
		}
		else if ( !_wcsicmp( Filename, L"clip:" ) ) {
			CloseMe			= FALSE;
			BOOL ClipError	= FALSE;

			if ( OpenClipboard( NULL ) ) {
				ClipHandle = GetClipboardData( CF_UNICODETEXT );
				if ( ClipHandle ) {
					ClipSize	= GlobalSize( ClipHandle );
					ClipTmp		= (BYTE*) GlobalLock( ClipHandle );

					if ( ClipSize && ClipTmp ) {

						if ( ClipboardData ) {
							free( ClipboardData );
							ClipboardData = NULL;
						}

						ClipboardData = (BYTE*) malloc( ClipSize );
						if ( ClipboardData ) {
							memcpy( ClipboardData, ClipTmp, ClipSize );
							ClipboardSize	= ClipSize;
							ClipPtr			= 0;
						}
						else {
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L"Couldn\u2019t allocate memory for clipboard data\r\n" );
							ClipError = TRUE;
						}
					}
					else {
						GetErrorText( GetLastError(), ErrorText );
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Failed to get clipboard data: \"%s\"\r\n", ErrorText );
						ClipError = TRUE;
					}

					if ( ClipTmp )
						GlobalUnlock( ClipHandle );
				}
				else {
					CloseClipboard();
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"No text on clipboard\r\n", ErrorText );
					ClipError = TRUE;
				}

				CloseClipboard();
			}
			else {
				GetErrorText( GetLastError(), ErrorText );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Error opening clipboard: \"%s\"\r\n", ErrorText );
				ClipError = TRUE;
			}

			if ( ClipError )
				return -1;
		}
		else {
			QueryTrueName( Filename, FilenameSave );

			FileHandle = CreateFile( FilenameSave, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL );

			if ( FileHandle == INVALID_HANDLE_VALUE ) {
				GetErrorText( GetLastError(), ErrorText );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Error opening file \"%s\":\r\n", FilenameSave );
				Qprintf( ERRHANDLE, L"%s\r\n", ErrorText );
				return -1;
			}

			CloseMe = TRUE;
		}


		//  Read in first bufferful:

		if ( ClipboardData )
			;
		else if ( ReadFile( FileHandle, FileBuffer, BUFFER_SIZE, &BytesRead, NULL ) == 0 ) {
			GetErrorText( GetLastError(), ErrorText );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error reading \"%s\" (1):\r\n", FilenameSave );
			Qprintf( ERRHANDLE, L"%s\r\n", ErrorText );
			CloseHandle( FileHandle );
			FileHandle = NULL;
			FileEOF = TRUE;
			BytesRead = BufferIndex = 0;

			if ( CloseMe )
				CloseHandle( FileHandle );
				
			FileHandle = NULL;
			return -1;
		}
		else if ( BytesRead == 0 ) {

			if ( CloseMe )
				CloseHandle( FileHandle );

			FileHandle = NULL;
			FileEOF = TRUE;
			BytesRead = BufferIndex = 0;
			return 0;
		}

		FileEncoding = FILE_OEM;

		if ( BytesRead >=2 )
			if ( IsTextUnicode( FileBuffer, BytesRead, NULL ) )
				FileEncoding = FILE_UTF16;

		if ( ClipboardData )
			FileEncoding = FILE_UTF16;

		if ( FileEncoding == FILE_UTF16 ) {
			if ( ( FileBuffer[0] == 0xff ) && ( FileBuffer[1] == 0xfe ) )
				BufferIndex = 2;
		}
		else if ( ( FileBuffer[0] == 0xef ) && ( FileBuffer[1] == 0xbb ) && ( FileBuffer[2] == 0xbf ) ) {
			BufferIndex = 3;
			FileEncoding = FILE_UTF8;
		}
	}


	do {
		lastch = ch;

		if ( !LiteralText )
			ch = GetChar();
		else {
			ch = LiteralText[i++];
			if ( ch == '\0' )
				FileEOF = TRUE;
		}

		skipme = FALSE;

		if ( ( j > 0 )  && ( ch < 0 ) )
			ch = '\0';

		if ( ch < 0 )
			break;						// abort on error or eof

		if ( ( ch == PILCROW ) | ( ch == SECTION_MARK ) )
			ch = '\0';

		if ( ch == CR ) {
			crcount++;
			ch = SPACE;
		}
		else if ( ch == LF ) {
			lfcount++;
			ch = SPACE;
		}
		else
			crcount = lfcount = 0;

		if ( ch == TAB )
			ch = SPACE;
		
		if ( ( ch == SPACE ) && ( j == 0 ) )
			skipme = TRUE;		// ignore leading spaces

		if ( ( ch == SPACE ) && ( lastch == SPACE ) )
			skipme = TRUE;		// ignore repeated spaces

		if ( !skipme && ( j < ( MAX_ARG_LEN - 1 ) ) ) {
			OutBuffer[j++] = ch;
			OutBuffer[j] = '\0';
		}

		if ( ch == '.' )
			dotscount++;

		if ( IsVowel( ch ) )
			vowelscount++;
		
		if ( ( crcount > ( BreakOnEOL ? 0 : 1 ) ) || ( lfcount > ( BreakOnEOL ? 0 : 1 ) )	// end of a paragraph
				|| ( ch == '\0' )									// null in the character stream
				|| ( ( ( ch == SPACE ) || ( ch == '"' ) ) && ( ( lastch == '!' ) || ( lastch == '?' ) ) )	// end of a sentence?
				|| ( ( ch == '"' ) && ( lastch == '.' ) )				// end of a sentence?
				|| ( ( ch == SPACE ) && ( lastch == '.' ) && ( dotscount < 2 ) && ( vowelscount > 0 ) ) // end of a sentence?
				|| ( ( ch == SPACE ) && ( j > ( MAX_ARG_LEN - 20 ) ) )	// buffer almost full
				|| ( j >= MAX_ARG_LEN )									// buffer completely full
				|| FileEOF ) {											// or nothing left to say

			//  speak the current line buffer and reset it

			if ( BreakCheck() ) {			// opportunity to abort with ctrl-break!

				if ( CloseMe )
					CloseHandle( FileHandle );

				if ( FileBuffer ) {
					FreeMem( FileBuffer );
					FileBuffer = NULL;
				}

				FileHandle = NULL;

				if ( ClipboardData ) {
					free( ClipboardData );
					ClipboardData	= NULL;
					ClipboardSize	= 0;
					ClipPtr			= 0;
				}

				return -3;
			}

			if ( j ) {

				if ( !LiteralText ) {
					utterance++;
					if ( Verbose )
						Printf( L"%04d: %s\r\n", utterance, OutBuffer );
				}
				else {
					CmdUtterance++;
					if ( Verbose )
						Printf( L"A%03d: %s\r\n", CmdUtterance, OutBuffer );
				}

				if ( SayLineNumbers ) {
					Sprintf( LineNumber, L"Line %d. ", utterance );
					MyVoice->Speak( LineNumber, SPF_DEFAULT, NULL );
					
					if ( BreakCheck() ) {

						if ( CloseMe )
							CloseHandle( FileHandle );

						if ( FileBuffer ) {
							FreeMem( FileBuffer );
							FileBuffer = NULL;
						}

						FileHandle = NULL;

						if ( ClipboardData ) {
							free( ClipboardData );
							ClipboardData	= NULL;
							ClipboardSize	= 0;
							ClipPtr			= 0;
						}

						return -3;			// opportunity to abort with ctrl-break!
					}
				}

				MyVoice->Speak( OutBuffer, SpeakFlags, NULL );
			}

			j = 0;
			OutBuffer[j] = '\0';
			crcount = 0;
			lfcount = 0;
			dotscount = 0;
			vowelscount = 0;
		}

		if ( ch == SPACE ) {
			dotscount = 0;
			vowelscount = 0;
		}

	} while ( !FileEOF );


	if ( CloseMe )
		CloseHandle( FileHandle );

	FileHandle = NULL;

	if ( ClipboardData ) {
		free( ClipboardData );
		ClipboardData	= NULL;
		ClipboardSize	= 0;
		ClipPtr			= 0;
	}

	return 0;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	if ( !HelpText )
		return 1;

	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



int SpeakSayHelp()
{
	RestartVoice();
	MyVoice->Speak( SaySpokenHelp, SPF_ASYNC | SPF_IS_XML, NULL );
	return 0;
}



DLLExports int WINAPI say( LPTSTR inString )
{
	if ( !inString )
		return 1;


	HRESULT rv = NULL;
	wchar_t outBuffer[MAX_ARG_LEN] = L"";
	int i = 0, j = 0, srv = 0, ReturnCode = 0;
	CmdUtterance = 0;
	BOOL wordstart = TRUE, quote = FALSE, filename = FALSE;
	wchar_t ch = '\0', swch = '\0';
	BOOL ValidAction     = FALSE;

	BOOL Verbose         = FALSE;
	BOOL SayLineNumbers  = FALSE;
	BOOL BreakOnEOL      = FALSE;
	BOOL EnableXML       = FALSE;
	BOOL Async           = FALSE;


	if ( !MyVoice ) {
		rv = CreateVoice();

		if ( rv < 0 ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle(STD_ERROR_HANDLE), L"CreateVoice() returned %d \r\n", rv );
			return 2;
		}
	}

	do {
		ch = inString[i++];

		if ( iswspace( ch ) && ( j == 0 ) )
			continue;

		if ( ( ch == '@' ) && wordstart && !quote && inString[i] && !iswspace( inString[i] ) ) {

			if ( j ) {
				if ( BreakCheck() ) {			// opportunity to abort with ctrl-break!

					if ( FileBuffer ) {
						FreeMem( FileBuffer );
						FileBuffer = NULL;
					}

					return 3;
				}

				srv = SpeakText( NULL, Verbose, FALSE, EnableXML, Async, FALSE, outBuffer );

				if ( srv == -3 )
					return 3;
				else if ( srv )
					ReturnCode = 2;
			}

			filename = TRUE;
			ValidAction = TRUE;
			j = 0;
			outBuffer[j] = '\0';
		}
		else if ( ( ch == '/' ) && wordstart && !quote )
			if ( inString[i] != '\0' )
				if ( iswspace( inString[i + 1] ) || ( inString[i + 1] == '\0' ) || ( inString[i + 1] == '/' ) ) {

					//  parse option:
					switch ( towupper( inString[i++] ) ) {

						case '?' :
							return ShowCmdHelp( SayHelpText, FALSE );

						case 'H' :
							SpeakSayHelp();
							return 0;

						case 'V' :
							Verbose = TRUE;
							continue;

						case 'X' :
							EnableXML = TRUE;
							continue;

						case 'D' :
							EnableXML = FALSE;
							continue;

						case 'A' :
							Async = TRUE;
							continue;

						case 'L' :
							BreakOnEOL = TRUE;
							continue;

						case 'N' :
							SayLineNumbers = TRUE;
							continue;

						case 'K' :
							RestartVoice();
							ValidAction = TRUE;
							continue;

						default :
							DisplayErrorHeader();
							Qprintf( GetStdHandle(STD_ERROR_HANDLE), L"Unknown option \"/%c\" \r\n", inString[i-1] );
							return 1;
					}
				}
				else if ( ( inString[i] == '?' ) && ( inString[i + 1] == '?' ) && ( iswspace( inString[i + 2] ) || ( inString[i + 2] == '\0' ) || ( inString[i + 2] == '/' ) ) )
					return ShowCmdHelp( SayHelpText, TRUE );


		if ( ch == '"' )
			quote = !quote;

		if ( ( ch == '"' ) && filename )
			continue;


		if ( ( iswspace( ch ) || ( ch == '\0' ) ) && filename && !quote ) {
			if ( j == 1 ) {
				filename = FALSE;
				continue;
			}
			else {
				// filename to read aloud:

				srv = SpeakText( outBuffer + 1, Verbose, SayLineNumbers, EnableXML, FALSE, BreakOnEOL, NULL );

				if ( srv == -3 )
					return 3;
				else if ( srv )
					ReturnCode = 2;

				ValidAction = TRUE;
				filename = FALSE;
				wordstart = TRUE;
				j = 0;

				if ( iswspace( ch ) )
					continue;
			}
		}

		outBuffer[j++] = ch;
		outBuffer[j] = '\0';
		wordstart = iswspace(ch);

	} while ( ch );


	if ( outBuffer[0] ) {
		if ( BreakCheck() ) {			// opportunity to abort with ctrl-break!

			if ( FileBuffer ) {
				FreeMem( FileBuffer );
				FileBuffer = NULL;
			}

			return 3;
		}

		srv = SpeakText( NULL, Verbose, FALSE, EnableXML, Async, FALSE, outBuffer );

		if ( srv == -3 )
			return 3;
		else if ( srv )
			ReturnCode = 2;

		ValidAction = TRUE;
	}

	if ( QueryIsConsole( GetStdHandle(STD_INPUT_HANDLE) ) == 0 ) {
		srv = SpeakText( L"-", Verbose, SayLineNumbers, EnableXML, FALSE, BreakOnEOL, NULL );

		if ( srv == -3 )
			return 3;
		else if ( srv )
			ReturnCode = 2;

		ValidAction = TRUE;
	}

	if ( !ValidAction )
		ShowCmdHelp( SayHelpText, FALSE );

	if ( FileBuffer ) {
		FreeMem( FileBuffer );
		FileBuffer = NULL;
	}

	return ReturnCode;
}



int ListVoices( LPTSTR Wildspec, BOOL Aloud )
{
	HRESULT hr = S_OK;
	CComPtr<IEnumSpObjectTokens> VoicesList;

	ULONG NumVoices					= 0;
	ISpObjectToken *ThisVoice		= NULL;
	ISpObjectToken *MyVoiceToken	= NULL;
	LPTSTR Description				= NULL;
	wchar_t DefaultDescript[]		= L"(none)";
	LPTSTR ThisVoiceID				= NULL;
	LPTSTR MyVoiceID				= NULL;
	BOOL ThisIsMyVoice				= FALSE;
	BOOL Match						= FALSE;
	

	if ( !SUCCEEDED( SpEnumTokens( SPCAT_VOICES, NULL, NULL, &VoicesList ) ) )
		return -1;

	VoicesList->GetCount( &NumVoices );
	if ( NumVoices == 0 )
		return 0;

	if ( MyVoice ) {
		MyVoice->GetVoice( &MyVoiceToken );
		if ( MyVoiceToken ) {
			MyVoiceToken->GetId( &MyVoiceID );
			MyVoiceToken->Release();
			MyVoiceToken = NULL;
		}
	}


	Printf( L"\r\n" );

	for ( unsigned int i = 0; i < NumVoices; i++ ) {
		VoicesList->Next( 1, &ThisVoice, NULL );
		ThisIsMyVoice = Match = FALSE;

		if ( ThisVoice ) {
			Description = NULL;
			SpGetDescription( ThisVoice, &Description );
			if ( !Description )
				Description = DefaultDescript;

			ThisVoiceID = NULL;
			ThisVoice->GetId( &ThisVoiceID );

			if ( MyVoiceID && ThisVoiceID )
				if ( _wcsicmp( MyVoiceID, ThisVoiceID ) == 0 )
					ThisIsMyVoice = TRUE;

			if ( ThisVoiceID )
				CoTaskMemFree( ThisVoiceID );

			if ( !Wildspec )
				Match = TRUE;
			else if ( !WildcardComparison( Wildspec, Description, 0, 1 ) )
				Match = TRUE;

			if ( Match ) {
				Printf( L"%4u %c  %s\r\n", i, ThisIsMyVoice ? '*' : 0x20, Description );

				if ( Aloud ) {
					wchar_t buffer[MAX_ARG_LEN] = L"";
					wsprintf( buffer, L"Number %u %s is %s.\n", i, ThisIsMyVoice ? L", this voice," : L"", Description );
					MyVoice->Speak( buffer, SPF_ASYNC, NULL );
				}
			}

			ThisVoice->Release();
			ThisVoice = NULL;
		}
	}


	if ( MyVoiceID )
		CoTaskMemFree( MyVoiceID );

	return NumVoices;
}



DLLExports int WINAPI _voice( LPTSTR outString )
{
	if ( !outString )
		return 1;


	if ( !MyVoice ) {
		wcscpy_s( outString, MAX_ARG_LEN, L"(none)" );
		return 0;
	}

	wcscpy_s( outString, MAX_ARG_LEN, L"(unknown)" );


	ISpObjectToken *MyVoiceToken	= NULL;
	LPTSTR Description				= NULL;

	MyVoice->GetVoice( &MyVoiceToken );
	if ( MyVoiceToken ) {
		SpGetDescription( MyVoiceToken, &Description );
		MyVoiceToken->Release();
	}

	if ( Description )
		wcscpy_s( outString, MAX_ARG_LEN, Description );
	
	return 0;
}



DLLExports int WINAPI _voicenum( LPTSTR outString )
{
	if ( !outString )
		return 1;


	wcscpy_s( outString, MAX_ARG_LEN, L"-1" );

	if ( !MyVoice )
		return 0;


	HRESULT hr						= S_OK;
	CComPtr<IEnumSpObjectTokens>	VoicesList;

	ULONG NumVoices					= 0;
	ISpObjectToken *ThisVoice		= NULL;
	ISpObjectToken *MyVoiceToken	= NULL;
	LPTSTR ThisVoiceID				= NULL;
	LPTSTR MyVoiceID				= NULL;
	unsigned int VoiceIndex			= UINT_MAX;



	if ( !SUCCEEDED( SpEnumTokens( SPCAT_VOICES, NULL, NULL, &VoicesList ) ) )
		return 0;

	VoicesList->GetCount( &NumVoices );
	if ( !NumVoices )
		return 0;


	if ( MyVoice ) {
		MyVoice->GetVoice( &MyVoiceToken );
		if ( MyVoiceToken ) {
			MyVoiceToken->GetId( &MyVoiceID );
			MyVoiceToken->Release();
		}
	}

	if ( !MyVoiceID )
		return 0;


	for ( unsigned int i = 0; i < NumVoices; i++ ) {
		VoicesList->Next( 1, &ThisVoice, NULL );

		if ( ThisVoice ) {
			ThisVoice->GetId( &ThisVoiceID );

			if ( ThisVoiceID ) {
				if ( !_wcsicmp( MyVoiceID, ThisVoiceID ) )
					VoiceIndex = i;

				CoTaskMemFree( ThisVoiceID );
				ThisVoice = NULL;
			}
		}
	}
	

	if ( MyVoiceID )
		CoTaskMemFree( MyVoiceID );

	if ( VoiceIndex != UINT_MAX )
		wsprintf( outString, L"%u", VoiceIndex );

	return 0;
}



int UseSystemDefaultVoice()
{
	if ( UserVoice )
		UserVoice->Release();

	UserVoice = NULL;
	MyVoice->SetVoice( NULL );
	return 0;
}



int SelectVoice( LPTSTR VoiceName )
{
	if ( !VoiceName )
		return UseSystemDefaultVoice();
	else if ( VoiceName[0] == '\0' )
		return UseSystemDefaultVoice();

	if ( !MyVoice )
		return 1;


	HRESULT hr = S_OK;
	CComPtr<IEnumSpObjectTokens> VoicesList;

	ULONG NumVoices					= 0;
	ISpObjectToken *ThisVoice		= NULL;
	ISpObjectToken *MatchingVoice	= NULL;
	LPTSTR Description				= NULL;
	LPTSTR ThisVoiceID				= NULL;
	BOOL Match						= FALSE;
	unsigned int DesiredIndex		= UINT_MAX;

	int rv		= 0;


	if ( iswdigit( VoiceName[0] ) )
		ParseInt( VoiceName, &DesiredIndex, NULL );


	if ( !SUCCEEDED( SpEnumTokens( SPCAT_VOICES, NULL, NULL, &VoicesList ) ) )
		return 2;

	VoicesList->GetCount( &NumVoices );
	if ( NumVoices == 0 )
		return 3;

	for ( unsigned int i = 0; i < NumVoices; i++ ) {
		VoicesList->Next( 1, &ThisVoice, NULL );
		Match = FALSE;

		if ( ThisVoice ) {
			if ( i == DesiredIndex )
				Match = TRUE;
			else {
				Description = NULL;
				SpGetDescription( ThisVoice, &Description );

				if ( Description )
					if ( !WildcardComparison( VoiceName, Description, 0, 1 ) )
						Match = TRUE;
			}

			if ( Match && !MatchingVoice )
				MatchingVoice = ThisVoice;
			else
				ThisVoice->Release();

			ThisVoice = NULL;
		}
	}

	if ( !MatchingVoice )
		return 4;


	if ( UserVoice )
		UserVoice->Release();

	UserVoice = MatchingVoice;
	if ( !SUCCEEDED ( MyVoice->SetVoice( MatchingVoice ) ) ) {
		rv = 10;
		UserVoice->Release();
		UserVoice = NULL;
	}

	return rv;
}



int SpeakVoiceHelp()
{
	RestartVoice();
	MyVoice->Speak( VoiceSpokenHelp, SPF_ASYNC | SPF_IS_XML, NULL );
	return 0;
}



DLLExports int WINAPI voice( LPTSTR inString )
{
	if ( !inString )
		return 1;


	LPTSTR NewVoiceName			= NULL;
	LPTSTR Wildspec				= NULL;
	BOOL ChangeVoice			= FALSE;
	BOOL ListAloud				= FALSE;

	wchar_t MyVoiceName[SHORT_BUF_LEN]		= L"";


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_ONE_ARG_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcscmp( arg[i], L"/?" ) )
				return ShowCmdHelp( VoiceHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( VoiceHelpText, TRUE );
			else if ( !_wcsicmp( arg[i], L"/A" ) )
				ListAloud = TRUE;
			else if ( !_wcsicmp( arg[i], L"/D" ) ) {
				NewVoiceName = NULL;
				ChangeVoice = TRUE;
			}
			else if ( !_wcsicmp( arg[i], L"/H" ) ) {
				SpeakVoiceHelp();
				return 0;
			}
			else if ( !_wcsnicmp( arg[i], L"/M", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] )
					Wildspec = arg[i] + j;
				else
					Wildspec = NULL;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( !NewVoiceName ) {
				NewVoiceName = arg[i];
				ChangeVoice = TRUE;
			}
			else {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
				return 1;
			}
		}
	}

	if ( !ChangeVoice ) {
		ListVoices( Wildspec, ListAloud );
		return 0;
	}

	if ( SelectVoice( NewVoiceName ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error selecting voice \"%s\"\r\n", NewVoiceName );
		return 1;
	}

	_voice( MyVoiceName );
	Printf( L"Voice set to %s.\r\n", MyVoiceName );

	if ( ListAloud ) {
		wchar_t buffer[MAX_ARG_LEN] = L"";
		wsprintf( buffer, L"Voice set to %s.", MyVoiceName );
		MyVoice->Speak( buffer, SPF_ASYNC, NULL );
	}

	return 0;
}



DLLExports int WINAPI _numvoices( LPTSTR outString )
{
	if ( !outString )
		return 1;


	CComPtr<IEnumSpObjectTokens> VoicesList;
	ULONG NumVoices = 0;

	wcscpy_s( outString, MAX_ARG_LEN, L"-1" );

	if ( !SUCCEEDED( SpEnumTokens( SPCAT_VOICES, NULL, NULL, &VoicesList ) ) )
		return 0;

	VoicesList->GetCount( &NumVoices );

	wsprintf( outString, L"%u", NumVoices );
	return 0;
}



DLLExports int WINAPI f_voice( LPTSTR String )
{
	if ( !String )
		return 1;


	HRESULT hr = S_OK;
	CComPtr<IEnumSpObjectTokens> VoicesList;

	unsigned int Index			= 0;
	ULONG NumVoices				= 0;
	ISpObjectToken *ThisVoice	= NULL;
	LPTSTR Description			= NULL;


	if ( ParseInt( String, &Index, NULL ) ) {
		DisplayErrorHeader();
		_wcsupr_s( String, MAX_ARG_LEN );
		Qprintf( ERRHANDLE, L"Bad index: \"%s\"\r\n", String );
		return -1;
	}


	wcscpy_s( String, MAX_ARG_LEN, L"-1" );

	if ( !SUCCEEDED( SpEnumTokens( SPCAT_VOICES, NULL, NULL, &VoicesList ) ) )
		return 0;

	VoicesList->GetCount( &NumVoices );

	if ( Index >= NumVoices )
		return 0;

	for ( unsigned int i = 0; i < Index; i++ ) {
		VoicesList->Next( 1, &ThisVoice, NULL );

		if ( ThisVoice ) {
			ThisVoice->Release();
			ThisVoice = NULL;
		}
	}

	VoicesList->Next( 1, &ThisVoice, NULL );
	if ( ThisVoice ) {
		SpGetDescription( ThisVoice, &Description );
		if ( Description ) {
			String[0] = '\0';
			wcscpy_s( String, MAX_ARG_LEN, Description );
		}

		ThisVoice->Release();
		ThisVoice = NULL;
	}

	return 0;
}



const int NumHelpTexts							= 6;

const wchar_t HelpFeature[NumHelpTexts][11]		= { L"say", L"voice", L"_numvoices", L"_voice", L"_voicenum", L"@voice" };
const wchar_t *HelpText[NumHelpTexts]			= { SayHelpText, VoiceHelpText, _NumVoicesHelpText, _VoiceHelpText, _VoiceNumHelpText, F_VoiceHelpText };



#ifndef	NEWHELP_VERSION

LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}

#endif	//	NEWHELP_VERSION



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( ( inString[0] == '@' ) || ( inString[0] == '_' ) )
		return 0;

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			wcscpy_s( outString, SHORT_BUF_LEN, UsageText( HelpText[i] ) );
			return 1;
		}
	}

	return 0;
}



#ifndef NEWHELP_VERSION

//		Simpler MessageBox() help, useful for smaller plugins which don't merit a full HTML help system:

DLLExports int WINAPI Help( LPTSTR inString )
{
	wchar_t WindowTitle[SHORT_BUF_LEN]		= L"";


	wsprintf( WindowTitle, L"%s Plugin Help", PLUGIN_NAME );

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			MessageBox( GetForegroundWindow(), HelpText[i], WindowTitle, MB_OK );
			return 1;
		}
	}

	return 0;
}


#endif		//	NEWHELP_VERSION


/*

	1.2.0.3		2023-10-15
				Minor cosmetic tweaks to ShowCmdHelp().

	1.2.1.0		2023-11-22
				Added support for @clip:
				Do not show "Plugin not loaded" messages in transient
				or pipe shells.

*/
